OW_NS = {} -- namespace

_,OW_NS.class = UnitClass("PLAYER") -- Store player class (English so no localisation needed)

local OW_tcs1 = ""
local OW_tcs2 = ""

local OW_buffspell = ""  -- spell to cast 
local OW_bufftarget = "" -- target to cast on

local OW_lasttime = 0 -- tracks GCD
local OW_lastcheck = 0 -- tracks last buffreminder
local OW_checkduration = 5 -- time between buffreminders - varied in the code...
local OW_lastreminder = ""


-- Initialise list of targets - TARGET, RAID1-40, PARTY1-5
local OW_targetlist = {"TARGET"}
for x=1,40 do
  table.insert(OW_targetlist,"RAID" .. x)
end
for x=1,5 do
  table.insert(OW_targetlist,"PARTY" .. x)
end
local OW_tpos = 1 -- where in the targetlist we're upto

local OW_playername = UnitName("PLAYER")

-- Talent inspection - player name and talent table
local OW_inspectname = ""
local OW_inspecttime = 0
OW_inspectDB = {}

local OW_defertench = ""
local OW_MHExp = 0
local OW_OHExp = 0

local OW_failed = {} -- table of players/buffs which failed
local OW_lastbuffspell = ""  -- last spell 
local OW_lastbufftarget = "" -- last target

function OW_NS.getsitubuffnames()
  local sb,ssb = "",""
  for _,sp in pairs(OW_NS.situbuffs) do
    if GetSpellInfo(sp) then
      if sb ~= "" then
        sb = sb .. ","
      end
      sb = sb .. sp
    end
  end
  for _,sp in pairs(OW_NS.selfsitubuffs) do
    if GetSpellInfo(sp) then
      if ssb ~= "" then
        ssb = ssb .. ","
      end
      ssb = ssb .. sp
    end
  end
  return sb,ssb
end 

local function OW_getbuff(target,n)
  name,_,_,stack,_,duration,expiry,ismine = UnitBuff(target,n) 
  return name,stack,duration,expiry,ismine
end 

-- Checks for "buffname" on "target" (partial matches e.g. "Blessing" will find "Blessing of Auchindoun" 
-- If mybuff = "1" it must be YOUR buff
-- Returns name of buff found
local function OW_CheckBuff(buffname,target,mybuff)
  local tu = UnitName(target) or target -- (could be "PLAYER" or "YOURNAME"
  local ft = OW_failed[buffname .. tu]
  if ft then
    if GetTime() - ft > 300 then -- ignored buff/player for 5 mins
      OW_failed[buffname .. tu] = nil
    else
      return buffname -- pretend we've found it
    end
  end
	local i=1
	local testname,_,duration,expiry,ismine=OW_getbuff(target,1)	-- * buffs
	while (testname) do
		if (strfind(strupper(testname),strupper(buffname))) then
      if (mybuff == 0) or (ismine) then
			  return testname
			end
		end
		i=i+1
		testname,_,duration,expiry,ismine=OW_getbuff(target,i)
	end
	return ""
end

-- Checks "target" for "buffname"
-- Returns true if not found - and readies buff for casting.
local function OW_NeedBuff(buffname,target)
 if IsUsableSpell(buffname) then -- not every Paladin has BOK etc.
   local st,du,en = GetSpellCooldown(buffname)
   if st and st > 0 then
     -- on cooldown
   else
    if (OW_CheckBuff(buffname,target,0) ~= "") then
      return false
    else
      OW_buffspell = buffname
      OW_bufftarget = target
      return true
    end
   end
 end
end

-- Checks spells against known aspect list (to avoid boss buff/debuffs with similar names)
local function OW_IsAspect(spell)
  if (spell == OW_NS.aoth) or (spell == OW_NS.aotc) or (spell == OW_NS.aotp) or
     (spell == OW_NS.aotv) or (spell == OW_NS.aotb) or (spell == OW_NS.aotw) or
     (spell == OW_NS.aotm) then
    return true
  else
    return false
  end
end

-- Checks spells against known aura list (to avoid boss buff/debuffs with similar names)
local function OW_IsAura(spell)
  if (spell == OW_NS.crusaura) or (spell == OW_NS.devaura) or (spell == OW_NS.concaura) or
     (spell == OW_NS.fireaura) or (spell == OW_NS.frostaura) or (spell == OW_NS.shadowaura) or
     (spell == OW_NS.retaura) or (spell == OW_NS.sancaura) then
    return true
  else
    return false
  end
end

-- Checks for Blessings (to avoid things like "Blessing of Auchindoun")
local function OW_HasMyBlessing(target)
  if (OW_CheckBuff(OW_NS.bom,target,1) ~= "") or (OW_CheckBuff(OW_NS.bow,target,1) ~= "") or
     (OW_CheckBuff(OW_NS.bok,target,1) ~= "") or (OW_CheckBuff(OW_NS.bos,target,1) ~= "") then 
    return true
  else
    return false
  end
end

-- Checks to see if you're in the Preparation zone of a BG
local function OW_InRaidPrep()
  if (OW_CheckBuff("Preparation","PLAYER",0) ~= "") then
    return true
  else
    return false
  end
end

-- Checks to see if you're in a BG
local function OW_inbg()
  if select(2, IsInInstance()) == "pvp" then
    return true
  else
    return false
  end
end

-- Checks raidgroup players of class "class" to see if they need "gblessing" 
-- if more then 3 do, returns true and readies spells for casting
local function OW_CheckGB(gblessing,blessing,class)
  if (IsUsableSpell(gblessing) == 1) then
    local tc = 0
    local tr = 0
    local targetname = ""
    for x = 1, 40 do
      local target = "RAID" .. x
      if UnitExists(target) then
        local _,cl = UnitClass(target)
        if (cl == class) then
          tc = tc + 1
          if (not UnitIsDeadOrGhost(target)) and (IsSpellInRange(gblessing,target) == 1) and (OW_CheckBuff(blessing,target,0) == "") then
            targetname = target
            tr = tr + 1
          end
        end
      end
    end
    -- We have 4-or-more buffable - and we're not missing more than 3 (e.g. we won't cast the greater buff twice!)
    if tr > 3 then
      OW_buffspell =  gblessing
      OW_bufftarget = targetname
      return true
    end
  end
  return false
end

-- Checks members of party "partyno" to see if they need have "buff" or "buff2" 
-- if more then 2 don't, returns true and readies buff for casting
local function OW_CheckParty(partyno,classes,notclasses,buff,buff2)
  if (IsUsableSpell(buff) == 1) then
    local tr = 0
    local targetname = ""
    if (GetNumRaidMembers() == 0) then
      if (partyno == 1) then -- not in raid, only handle Party1 
        for x = 1,5 do
          local target = "PARTY" .. x
          if UnitExists(target) then
            local _,cl = UnitClass(target) -- non-localised classname
            if (classes == "" or strfind(classes,cl)) or (notclasses == "" or not strfind(notclasses,cl)) then
              if (not UnitIsDeadOrGhost(target)) and IsSpellInRange(buff,target) == 1 and 
                  OW_CheckBuff(buff,target,0) == "" and (not buff2 or OW_CheckBuff(buff2,target,0) == "") then
                targetname = UnitName(target)
                tr = tr + 1
              end
            end
          end
        end
      end
    else
      for x = 1,GetNumRaidMembers() do
        local name,_,subg = GetRaidRosterInfo(x)
        --if subg == partyno then Spells no longer party specific
          local target = name
          if UnitExists(target) then
            local _,cl = UnitClass(target) -- non-localised classname
            if (classes == "" or strfind(classes,cl)) or (notclasses == "" or not strfind(notclasses,cl)) then
              if (not UnitIsDeadOrGhost(target)) and (IsSpellInRange(buff,target) == 1) and 
                  OW_CheckBuff(buff,target,0) == "" and (not buff2 or OW_CheckBuff(buff2,target,0) == "") then
                targetname = name
                tr = tr + 1
              end
            end
          end
        --end
      end
    end
    if tr > 3 then
      OW_buffspell =  buff
      OW_bufftarget = targetname
      return true
    end
  end
  return false
end

-- This function will set OW_buffspell and OW_bufftarget to the next buff required
-- It's used by the reminder AND casting local functions
local function OW_SetSpellCast()
  OW_buffspell = ""
  OW_bufftarget = ""
  -- If GCD not passed since last spell cast - do nothing
  if (GetTime() - OW_lasttime > 1.6) then

    -- Group Buffs first so we avoid self-buffing and THEN group buffing  
    -- Start Priest/Druid/Mage Group Buffs
    if (not IsMouseButtonDown("LeftButton")) then
      if (GetNumPartyMembers() == 0) then
        ng = 0
      elseif (GetNumRaidMembers() == 0) then
        ng = 1
      else
        ng = 8 -- even in a 2-man raid people can be in Group 8
      end
      --for gn = 1,ng do  parties no long apply to spells
        if OW_NS.class == "PRIEST" then
          if OW_CheckParty(ng,"","",OW_NS.pof,OW_NS.pwf) then return true end
          if OW_CheckBuff(OW_NS.prayersprot,"PLAYER",0) ~= "" or OW_CheckBuff(OW_NS.sprot,"PLAYER",0) ~= "" then
            if OW_CheckParty(ng,"","",OW_NS.prayersprot,OW_NS.sprot) then return true end
          end
          if OW_CheckParty(ng,"MAGE;PRIEST;SHAMAN;DRUID;PALADIN","",OW_NS.prayerspirit,OW_NS.divinespirit) then return true end
        elseif OW_NS.class == "DRUID" then
          if OW_CheckParty(ng,"","",OW_NS.gotw,OW_NS.motw) then return true end
        elseif OW_NS.class == "MAGE" then
          if OW_CheckParty(ng,"","WARRIOR;ROGUE;DEATHKNIGHT",OW_NS.ab,OW_NS.ai) then return true end
        end
      --end
    end
    -- End Priest/Druid/Mage Group Buffs

    -- Paladin Self-buffs
    if OW_NS.class == "PALADIN" then
      if (IsMounted("PLAYER")) then
        -- Apply Crusader
        if (OW_NeedBuff(OW_NS.crusaura,"PLAYER") == true) then
          OW_tcs2 = "/stopcasting\n/cast [target=PLAYER] " .. OW_spell1
          return true
        else 
          return false
        end
      else
        -- Reapply last aura saved when mounted (and then clear it)
        if (OW_spell1) and (OW_spell1 ~= "") then
          if (OW_NeedBuff(OW_spell1,"PLAYER") == true) then
            OW_tcs2 = ""
            return true
          end
        end
        -- Blessings
        if (OW_HasMyBlessing("PLAYER") == false) then  -- if we don't have one...
          if OW_spell2 and OW_spell2 ~= "" then -- what was our last blessing?
            if (OW_NeedBuff(OW_spell2,"PLAYER") == true) then return true end  -- cast it again
          else
            if IsEquippedItemType(OW_NS.Twohand) then -- 2H = Might
              if (OW_NeedBuff(OW_NS.bom,"PLAYER") == true) then return true end
            --  Blessing of Kings
            elseif (OW_NeedBuff(OW_NS.bok,"PLAYER") == true) then return true
            --  still no Blessing? - Blessing of Wisdom then!
            elseif (OW_HasMyBlessing("PLAYER") == false) then -- No self-Blessing
              if (OW_NeedBuff(OW_NS.bow,"PLAYER") == true) then return true end
            end
          end
        end

      end -- mounted/non-mounted
    end
    -- End Paladin Self-buffs

    -- Priest Self-buffs
    if OW_NS.class == "PRIEST" then
      if (not IsMounted("PLAYER")) then
        -- Power Word: Fortitude
        if (OW_CheckBuff(OW_NS.pof,"PLAYER",0) == "") and (OW_NeedBuff(OW_NS.pwf,"PLAYER") == true) then return true end
        -- Divine Spirit
        if ((OW_CheckBuff(OW_NS.prayerspirit,"PLAYER",0) == "") and (OW_NeedBuff(OW_NS.divinespirit,"PLAYER") == true)) then return true end
      end
    end
    -- End Priest Self-buffs

    -- Mage Self-buffs
    if OW_NS.class == "MAGE" then
      if (not IsMounted("PLAYER")) then
        -- Arcane Intellect
        if (OW_CheckBuff(OW_NS.ab,"PLAYER",0) == "") and (OW_NeedBuff(OW_NS.ai,"PLAYER") == true) then return true end
        if (OW_CheckBuff(OW_NS.icearmor,"PLAYER",1) == "") and (OW_CheckBuff(OW_NS.magearmor,"PLAYER",1) == "") and 
           (OW_CheckBuff(OW_NS.frostarmor,"PLAYER",1) == "") and (OW_CheckBuff(OW_NS.moltenarmor,"PLAYER",1) == "") then 
          if OW_spell2 and OW_spell2 ~= "" then -- what was our last armor?
            if (OW_NeedBuff(OW_spell2,"PLAYER") == true) then return true end  -- cast it again
          end
        end
      end
    end
    -- End Mage Self-buffs

    -- Hunter Self-buffs
    if OW_NS.class == "HUNTER" then
      if (not IsMounted("PLAYER")) then
        -- Trueshot Aura
        if (OW_NeedBuff(OW_NS.tsa,"PLAYER") == true) then return true end
        -- If out-of-combat and no runspeed aspect...
        if (OW_CheckBuff(OW_NS.aotc,"PLAYER",1) == "") and (OW_CheckBuff(OW_NS.aotp,"PLAYER",1) == "") and (not UnitAffectingCombat("PLAYER")) then
          OW_tcs2 = "/stopcasting\n/cast [target=PLAYER] " .. OW_spell1
          -- Pack if in party and NOT Battleground - Cheetah otherwise
          if (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and (not OW_inbg) then
            if (OW_NeedBuff(OW_NS.aotp,"PLAYER") == true) then return true end
          else
            if (OW_NeedBuff(OW_NS.aotc,"PLAYER") == true) then return true end
          end
        else
          OW_tcs2 = ""
          if (OW_spell1) and (OW_spell1 ~= "") then
            -- Switch to last non-runspeed Aspect
            if (OW_NeedBuff(OW_spell1,"PLAYER") == true) then
              return true
            end
          end
        end
      end
    end
    -- End Hunter Self-buffs

    -- Warlock Self-buffs
    if OW_NS.class == "WARLOCK" then
      if (not IsMounted("PLAYER")) then
        -- If armor active
        if (OW_CheckBuff(OW_NS.demonskin,"PLAYER",1) ~= "") or (OW_CheckBuff(OW_NS.demonarmor,"PLAYER",1) ~= "") or (OW_CheckBuff(OW_NS.felarmor,"PLAYER",1) ~= "") then
        elseif OW_spell2 and OW_spell2 ~= "" then
          -- cast last armour
          if (OW_NeedBuff(OW_spell2,"PLAYER") == true) then return true end
        end
      end
    end
    -- End Warlock Self-buffs

    -- Druid Self-buffs
    if OW_NS.class == "DRUID" then
      if (not IsMounted("PLAYER")) then 
        -- Mark of the Wild
        if ((OW_CheckBuff(OW_NS.gotw,"PLAYER",0) == "") and (OW_NeedBuff(OW_NS.motw,"PLAYER") == true)) then return true end
        -- Omen of Clarity
        if (IsUsableSpell(OW_NS.omen)) then
          if (OW_NeedBuff(OW_NS.omen,"PLAYER") == true) then return true end
        end
      end
    end
    -- End Druid Self-buffs

    -- Shaman/Rogue Self-buffs (identical in terms of MH/OH buffing)
    if OW_NS.class == "ROGUE" or OW_NS.class == "SHAMAN" then
      if (not IsMounted("PLAYER")) then 
        local mhw = GetInventoryItemLink("PLAYER",16)
        local ohw = GetInventoryItemLink("PLAYER",17)
        local mh,_,_,oh = GetWeaponEnchantInfo()
        if mhw and (not mh) and OW_tench1 and OW_tench1 ~= "" then
          OW_buffspell = OW_tench1 .. OW_NS.mh
          OW_bufftarget = "PLAYER"
          return true
        elseif not IsEquippedItemType(OW_NS.Shields) and ohw and (not oh) and OW_tench2 and OW_tench2 ~= "" then
          OW_buffspell = OW_tench2 .. OW_NS.oh
          OW_bufftarget = "PLAYER"
          return true
        end
      end
    end
    -- End Shaman/Rogue Self-buffs

    -- Situational Buffs
    if (not IsMounted("PLAYER")) then 
      for target in pairs(OW_NS.situbuffdb) do
        for spell,ct in pairs(OW_NS.situbuffdb[target]) do
          -- this will fail if someone left party/raid so no need for UnitExists
          if target == OW_playername or IsSpellInRange(spell,target) == 1 then
            if (OW_NeedBuff(spell,target) == true) then return true end
          end
        end
      end
    end
    
    -- After self-buffs because we never buff ourselves here...
    -- Start Paladin Group Buffs
    if (not IsMouseButtonDown("LeftButton")) then
      if OW_NS.class == "PALADIN" then
        if OW_CheckGB(OW_NS.gbom,OW_NS.bom,"DEATHKNIGHT") then return true end
        if OW_CheckGB(OW_NS.gbom,OW_NS.bom,"WARRIOR") then return true end
        if OW_CheckGB(OW_NS.gbom,OW_NS.bom,"ROGUE") then return true end
        if OW_CheckGB(OW_NS.gbom,OW_NS.bom,"HUNTER") then return true end
        if OW_CheckGB(OW_NS.gbow,OW_NS.bow,"MAGE") then return true end
        if OW_CheckGB(OW_NS.gbow,OW_NS.bow,"PRIEST") then return true end
        if OW_CheckGB(OW_NS.gbok,OW_NS.bok,"WARLOCK") then return true end
      end
    end
    -- End Paladin Group Buffs
  
    -- Now we go through the targetlist from where we last left it (to avoid getting stuck on people out of line-of-sight)
    local stpos = OW_tpos
    while 1 == 1 do -- yes, loop endlessly - breakout once we've been through the entire list (see end of loop) 
      local target = OW_targetlist[OW_tpos]
      OW_tpos = OW_tpos + 1
      if (OW_tpos > # OW_targetlist) then 
        OW_tpos = 1
      end
      
      -- Check our chosen target exists, is a player, is friendly, is alive, and is not disconnected
      if (UnitExists(target)) and (UnitIsPlayer(target)) and (UnitIsFriend("PLAYER",target)) and (not UnitIsDeadOrGhost(target)) and (UnitIsConnected(target)) then
        class,eclass = UnitClass(target) -- ENGLISH classname so no localisation needed
              
        -- Handle Paladin buffs for players in range of Blessings...
        if (OW_NS.class == "PALADIN") and (IsSpellInRange(OW_NS.bom,target) == 1) then
          if (not IsMounted("PLAYER")) then
            if (UnitIsUnit(target,"PLAYER")) then
              -- ignore self 
            elseif (OW_HasMyBlessing(target)) then
              -- player already has our blessing
              -- if it's a hunter, we check pet
              if (UnitExists(target .. "PET")) then  
                 if (OW_HasMyBlessing(target .. "PET")) then
                    --pet already has our blessing
                 elseif (eclass == "HUNTER") then
                    if (OW_NeedBuff(OW_NS.bom,target .. "PET") == true) then return true end
                 elseif (eclass == "WARLOCK") then
                    if (OW_NeedBuff(OW_NS.bok,target .. "PET") == true) then return true end
                 end
              end
            else
              -- Shamans and Druids we fire an Inspect request to check talents...
              if (eclass == "SHAMAN") or (eclass == "DRUID") then
                local unitname = UnitName(target)
                if ((OW_inspectname ~= "") and (GetTime() - OW_inspecttime < 5)) then 
                    -- already awaiting inspect data - skip player for now
                elseif (not OW_inspectDB[unitname]) then -- player not in talent table
                   -- send inspect request
                   OW_inspectname = unitname
                   OW_inspecttime = GetTime()
                   NotifyInspect(target)
                else
                   local spec = OW_inspectDB[unitname]
                   -- Resto Shamans/Druids - Blessing of Wisdom
                   if ((eclass == "DRUID") and (spec[3] > 30)) or
                      ((eclass == "SHAMAN") and (spec[3] > 30)) then 
                        if (OW_NeedBuff(OW_NS.bow,target) == true) then return true end
                   end
                   -- Feral Druids/Enhancement Shamans - Blessing of Might
                   if ((eclass == "DRUID") and (spec[2] > 30)) or
                      ((eclass == "SHAMAN") and (spec[2] > 30)) then 
                        if (OW_NeedBuff(OW_NS.bom,target) == true) then return true end
                   end
                   -- All Druids/Shamans we've not managed to buff so far - try Kings, Wisdom, Might
                   if (OW_NeedBuff(OW_NS.bok,target) == true) then return true end
                   if (OW_NeedBuff(OW_NS.bow,target) == true) then return true end
                   if (OW_NeedBuff(OW_NS.bom,target) == true) then return true end
                end
              else                
                -- Hunters, Rogues, Warriors, DeathKnights - Blessing of Might
                if ((eclass == "WARRIOR") or (eclass == "ROGUE") or (eclass == "HUNTER") or (eclass == "DEATHKNIGHT")) then
                  if (OW_NeedBuff(OW_NS.bom,target) == true) then return true end
                end
                -- Mages, Priests - Blessing of Wisdom
                if (eclass == "MAGE") or (eclass == "PRIEST") then
                  if (OW_NeedBuff(OW_NS.bow,target) == true) then return true end
                end
                -- Everyone we've not managed to buff so far - Blessing of Kings
                if (OW_NeedBuff(OW_NS.bok,target) == true) then return true end
                -- Hunters not blessed so far - Blessing of Wisdom
                if (eclass == "HUNTER") then
                  if (OW_NeedBuff(OW_NS.bow,target) == true) then return true end
                end
                -- Paladins not blessed so far - Blessing of Might
                if (eclass == "PALADIN") then
                  if (OW_NeedBuff(OW_NS.bom,target) == true) then return true end
                end
              end
            end
          end
        end
        -- End Paladin Buffs  

        -- Handle Priest buffs for players in range of PWF...
        if (OW_NS.class == "PRIEST") and (IsSpellInRange(OW_NS.pwf,target) == 1) then
          if (not IsMounted("PLAYER")) then
            if (UnitIsUnit(target,"PLAYER")) then
              -- ignore self
            -- Power Word: Fortitude
            elseif (OW_CheckBuff(OW_NS.pof,target,0) == "") and (OW_NeedBuff(OW_NS.pwf,target) == true) then return true
            -- Power Word: Fortitude on Pets
            elseif eclass ~="SHAMAN" and UnitExists(target .."PET") and
                   OW_NeedBuff(OW_NS.pwf,target .. "PET") == true then return true
            -- Divine Spirit on Healers/Mages (seems excessive to check for 'resto' shamans/druids here)
            elseif ((eclass == "PRIEST") or (eclass == "MAGE") or (eclass == "DRUID") or (eclass == "SHAMAN") or (eclass == "PALADIN")) and
                   OW_CheckBuff(OW_NS.prayerspirit,target,0) == "" and OW_NeedBuff(OW_NS.divinespirit,target) == true then return true
            -- Shadow Prot on all players IF the player already has it
            elseif (OW_CheckBuff(OW_NS.prayersprot,"PLAYER",0) ~= "" or OW_CheckBuff(OW_NS.sprot,"PLAYER",0) ~= "") and
                   OW_CheckBuff(OW_NS.prayersprot,target,0) == "" and OW_NeedBuff(OW_NS.sprot,target) == true then return true 
            end
          end
        end
        -- End Priest Buffs  

        -- Handle Mage buffs for players in range of AI...
        if (OW_NS.class == "MAGE") and (IsSpellInRange(OW_NS.ai,target) == 1) then
          if (not IsMounted("PLAYER")) then
            if (UnitIsUnit(target,"PLAYER")) then
              -- ignore self
            -- Arcane Intellect
            elseif (eclass ~= "WARRIOR") and (eclass ~= "ROGUE") and (eclass ~= "DEATHKNIGHT")  then -- non-mana classes 
              if (OW_CheckBuff(OW_NS.ab,target,0) == "") and (OW_NeedBuff(OW_NS.ai,target) == true) then return true end
            elseif eclass == "WARLOCK" and UnitExists(target .."PET") then
              if OW_NeedBuff(OW_NS.ai,target.."PET") == true then return true end
            end
          end
        end
        -- End Mage Buffs  

        -- Handle Druid buffs for players in range of MOTW...
        if (OW_NS.class == "DRUID") and (IsSpellInRange(OW_NS.motw,target) == 1) then 
          if (not IsMounted("PLAYER")) then  
            if (UnitIsUnit(target,"PLAYER")) then
              -- ignore self
            -- Mark of the Wild
            elseif ((OW_CheckBuff(OW_NS.gotw,target,0) == "") and (OW_NeedBuff(OW_NS.motw,target) == true)) then return true
            -- Mark of the Wild on Hunter Pets
            elseif (eclass == "HUNTER") and (UnitExists(target .."PET")) then
              if ((OW_CheckBuff(OW_NS.gotw,target .. "PET",0) == "") and (OW_NeedBuff(OW_NS.motw,target .. "PET") == true)) then return true end
            end
          end
        end
        -- End Druid Buffs  
  
      end -- unitexists etc.
      
      if (OW_tpos == stpos) then -- we've been through the whole list so breakout!
        return true -- allow default macros to cast EVEN if we're mounted...
      end
  
    end -- endless loop
  
  else -- in GCD
    return false
  end
end

-- This finds and displays any mousewheel-related bindings -
local function OW_checkbindings()
  for x = 1,GetNumBindings() do
    local action = select(1,GetBinding(x))
    for y = 2,select("#",GetBinding(x)) do
       local keyb = select(y,GetBinding(x))
       if strfind(strupper(keyb),"-MOUSEWHEEL") then
         DEFAULT_CHAT_FRAME:AddMessage("O-Wheely!: Function " .. action .. " is bound to " .. keyb,1,0,0)
       end
    end
  end
end

-- Returns either the OWUp/OWDown macro content - or preset combatspell1,2...
local function OW_GetMWDefaults()
  local mt,mt2
  if not OW_combatspell or OW_combatspell == "" then
    mt = OW_tcs1
  else
    mt = OW_combatspell
  end
  if not OW_combatspell2 or OW_combatspell2 == "" then
    mt2 = OW_tcs2
  else
    mt2 = OW_combatspell2
  end
  return mt,mt2
end

local function OW_init()
  --Savedvariables
  if not OW_options then
    OW_options = {}
    OW_spell1 = "" -- Aspects and Auras
    OW_spell2 = "" -- Blessings, Armors and Shields
    OW_tench1 = "" -- MH Temp Enchant
    OW_tench2 = "" -- OH Temp Enchant
    OW_selfsitu = {} -- stored self-situational buffs
  end

  OW_NS.situbuffs = {OW_NS.bosa, OW_NS.thorns, OW_NS.ubreath, OW_NS.dinvis, OW_NS.amplify, OW_NS.dampen}
  OW_NS.selfsitubuffs = {OW_NS.boneshield, OW_NS.righteous, OW_NS.lshield, OW_NS.wshield, OW_NS.innerfire, OW_NS.sprot}
  OW_NS.situbuffdb = {}
  OW_NS.situbuffdb[OW_playername] = OW_selfsitu

  OW_NS.optionsinit() -- initialise options screen

  OW_NS.bindings()

  OW_BuffFrame:Click() -- Initialise it

  _G['OWheely']:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

-- INSPECT_TALENT_READY event
-- Inspected Player's name in OW_inspectname
local function OW_CheckTalent()
  OW_inspectDB[OW_inspectname] = {}
  local tn1,_,t1 = GetTalentTabInfo(1,1)
  local tn2,_,t2 = GetTalentTabInfo(2,1)
  local tn3,_,t3 = GetTalentTabInfo(3,1)
  OW_inspectDB[OW_inspectname][1] = t1 or 0
  OW_inspectDB[OW_inspectname][2] = t2 or 0
  OW_inspectDB[OW_inspectname][3] = t3 or 0
  OW_inspectname = ""
end

-- COMBAT_LOG_EVENT_UNFILTERED event 
local function OW_Combat()
  if (arg2) and (arg2 == "SPELL_CAST_SUCCESS") and (arg4) and (arg4 == OW_playername) and arg9 then 
    if (not arg7) then
      arg7 = arg4 -- no target (arg7) means us (arg4)!
    end
    local spellname = GetSpellInfo(arg9)
    if (spellname) and (spellname ~= "") then -- valid spell?
      if (arg7 == arg4) then -- casting on ourselves 
        if (spellname == OW_NS.bok) or (spellname == OW_NS.bow) or (spellname == OW_NS.bom) or (spellname == OW_NS.bos) or
           (spellname == OW_NS.icearmor) or (spellname == OW_NS.frostarmor) or (spellname == OW_NS.magearmor) or (spellname == OW_NS.moltenarmor) or
           (spellname == OW_NS.demonskin) or (spellname == OW_NS.demonarmor) or (spellname == OW_NS.felarmor) then
              OW_spell2 = spellname
        end
        if (OW_IsAspect(spellname) and (spellname ~= OW_NS.aotc) and (spellname ~= OW_NS.aotp)) or
           (OW_IsAura(spellname) and (spellname ~= OW_NS.crusaura)) then
              OW_spell1 = spellname
        end
      end
      -- because we're using NAMES we need to check they're valid units
      -- only members of our party/raid will return true - if they leave party/raid we can't rebuff em
      if arg7 ~= OW_playername and UnitExists(arg7) then 
        for i,sn in pairs(OW_NS.situbuffs) do
          if spellname == sn then 
            if not OW_NS.situbuffdb[arg7] then
              OW_NS.situbuffdb[arg7] = {}
            end
            OW_NS.situbuffdb[arg7]{[sn]  = GetTime()}
          end
        end
      elseif arg7 == OW_playername then
        for i,sn in pairs(OW_NS.selfsitubuffs) do
          if spellname == sn then 
            OW_NS.situbuffdb[arg7][sn] = GetTime()
            OW_selfsitu = OW_NS.situbuffdb[arg7]
          end
        end
      end
    end
  elseif (arg2) and (arg2 == "ENCHANT_APPLIED") and (arg4) and (arg4 == OW_playername) and (arg9) then  
    local spellname = OW_NS.tenchtospell(arg9) -- translate the tenchname (no SpellId available) to the spellname 
    if spellname and 
       ((OW_NS.class == "ROGUE" and strfind(strupper(spellname),strupper(OW_NS.poison))) or
        (OW_NS.class == "SHAMAN" and spellname == OW_NS.rockbiter or spellname == OW_NS.flametongue or 
         spellname == OW_NS.frostbrand or spellname == OW_NS.windfury)) then
      local mhw = GetInventoryItemLink("PLAYER",16)
      local ohw = GetInventoryItemLink("PLAYER",17)
      if mhw and (strfind(mhw,arg11)) and ohw and (strfind(ohw,arg11)) then -- identical weapons
        -- GetWeaponEnchantInfo() will show which weapon was enchanted (longest time remaining) 
        -- BUT - it's not been updated yet - the server is slow about this...
        -- so we store the poisonname and expiry times and check them later (in OW_Update)
        OW_defertench = spellname
        _,OW_MHExp,_,_,OW_OHExp = GetWeaponEnchantInfo()
        if not OW_MHExp then OW_MHExp = 0 end
        if not OW_OHExp then OW_OHExp = 0 end
      elseif mhw and (strfind(mhw,arg11)) then
        OW_tench1 = spellname
      elseif ohw and (strfind(ohw,arg11)) then
        OW_tench2 = spellname
      end
    end
  elseif (arg2) and (arg2 == "SPELL_CAST_FAILED") and (arg4) and (arg4 == OW_playername) and (arg10) then
    -- "A more powerful spell" failures do NOT have a 'target' so we can't tell WHO they failed ON
    -- We check if it's the last buffspell cast and if it is, assume the target was our last bufftarget
    if OW_lastbuffspell and OW_lastbufftarget and arg10 == OW_lastbuffspell and strfind(strupper(arg12),strupper(OW_NS.morepowerful)) then 
      OW_failed[arg10 .. UnitName(OW_lastbufftarget)] = GetTime()
    end
  elseif (arg2) and (arg2 == "SPELL_AURA_REMOVED") and (arg7) and (arg7 == OW_playername) and (arg10) then
    sn = arg10
    if OW_NS.situbuffdb[OW_playername][sn] and GetTime() - OW_NS.situbuffdb[OW_playername][sn] < 6 then
      OW_NS.situbuffdb[OW_playername][sn] = nil
    end
  end
end

local function OW_event(self, event, arg1)
  if (event == "PLAYER_ENTERING_WORLD") then
     OW_init(); 
  elseif (event == "INSPECT_TALENT_READY") then 
     OW_CheckTalent(); 
  elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then 
     OW_Combat(); 
  end 
end

local function OW_buffreminder(reminder)
  if reminder == OW_lastreminder then 
    if OW_checkduration < 30 then -- actually maxes at about 38 seconds
      OW_checkduration = OW_checkduration + (OW_checkduration / 2)
    end
  end
  UIErrorsFrame:AddMessage(reminder,255,0,0)
  OW_lastreminder = reminder
end


-- ON_UPDATE event
function OW_update()
  -- we deferred a temp. enchant application as we're using identical weapons
  -- check to see if the temp. enchant expiry has updated (showing which weapon was enchanted)
  if (OW_defertench ~= "") then
    local _,mhexp,_,_,ohexp = GetWeaponEnchantInfo()
    if not mhexp then mhexp = 0 end
    if not ohexp then ohexp = 0 end
    if mhexp > OW_MHExp then -- shows we applied it to MH
      OW_tench1 = OW_defertench
      OW_defertench = ""
    elseif ohexp and OW_OHExp and ohexp > OW_OHExp then -- shows we applied it to OH
      OW_tench2 = OW_defertench
      OW_defertench = ""
    end
  end
  -- Out of combat - every 5 secs - check for buffs needed
  -- NOT when resting and NOT when in raid/not in a BG
  if (GetTime() - OW_lastcheck > OW_checkduration) and
     (not IsResting()) and
     ((GetNumRaidMembers() == 0) or (OW_inbg)) then
    OW_lastcheck = GetTime()
    OW_SetSpellCast()
    -- ignore Hunter and Paladin 'speed' aspect/auras
    if (OW_buffspell ~= "") and (OW_buffspell ~= OW_aotp) and (OW_buffspell ~= OW_aotc) then
      if ((not UnitAffectingCombat("PLAYER")) and (not OW_IsAspect(OW_buffspell))) or 
         ((UnitAffectingCombat("PLAYER")) and (OW_NS.class == "PALADIN") and OW_IsAura(OW_buffspell)) or
         ((UnitAffectingCombat("PLAYER")) and (OW_NS.class == "HUNTER") and OW_IsAspect(OW_buffspell)) then
          OW_buffreminder(OW_NS.needs(UnitName(OW_bufftarget),OW_buffspell))
      end
    end
  end
end

local OWheely = CreateFrame("Frame","OWheely")
OWheely:RegisterEvent("PLAYER_ENTERING_WORLD")
OWheely:RegisterEvent("INSPECT_TALENT_READY");
OWheely:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
OWheely:SetScript("OnEvent", OW_event)
OWheely:SetScript("OnUpdate",OW_update)

-- handles camera zooming 
local function OW_camera(mwd)
  if OW_options and OW_options['CZOOM'] then
  elseif IsMouseButtonDown("RightButton") then
    if mwd == "MWUP" then
      CameraZoomIn(1)
    elseif mwd == "MWDN" then
      CameraZoomOut(1)
    end
  end
end

local function OW_preclick(self)
  local ssc = OW_SetSpellCast()
  if (not UnitAffectingCombat("PLAYER")) then 
    if not ssc or (OW_options and not OW_options['CZOOM'] and IsMouseButtonDown("RightButton")) then 
  	  self:SetAttribute("type", "")
  	  self:SetAttribute("macrotext-mwup", "")
  	  self:SetAttribute("ctrl-macrotext-mwup", "")
  	  self:SetAttribute("alt-macrotext-mwup", "")
  	  self:SetAttribute("shift-macrotext-mwup", "")
  	  self:SetAttribute("macrotext-mwdn", "")
  	  self:SetAttribute("ctrl-macrotext-mwdn", "")
  	  self:SetAttribute("alt-macrotext-mwdn", "")
  	  self:SetAttribute("shift-macrotext-mwdn", "")
    else
      local mt,mt2 = ""
      if (OW_buffspell ~= "") and (((OW_NS.class ~= "DRUID") and (OW_NS.class ~= "SHAMAN")) or (GetShapeshiftForm() == 0)) then
         mt = "/cast [target=" .. OW_bufftarget .. "] " .. OW_buffspell
         mt = string.gsub(mt,OW_NS.mh,"%\n%/use 16") -- handles rogue/shaman temp enchants
         mt = string.gsub(mt,OW_NS.oh,"%\n%/use 17") -- handles rogue/shaman temp enchants
         mt2 = mt
         OW_lasttime = GetTime() -- trigger the GCD timer
         OW_lastbuffspell = OW_buffspell
         OW_lastbufftarget = OW_bufftarget
      else
         mt,mt2 = OW_GetMWDefaults()
      end
  	  self:SetAttribute("type", "macro")
  	  self:SetAttribute("macrotext-mwup", mt)
  	  self:SetAttribute("ctrl-macrotext-mwup", mt)
  	  self:SetAttribute("alt-macrotext-mwup", mt)
  	  self:SetAttribute("shift-macrotext-mwup", mt)
  	  self:SetAttribute("macrotext-mwdn", mt2)
  	  self:SetAttribute("ctrl-macrotext-mwdn", mt2)
  	  self:SetAttribute("alt-macrotext-mwdn", mt2)
  	  self:SetAttribute("shift-macrotext-mwdn", mt2)
    end
  end
end

local function OW_postclick(self,arg1)
  OW_checkduration = 5 -- resets the reminder timer
  OW_camera(arg1)
  -- self ensures that the 'default' spell will be active after combatlockdown...
  if (not UnitAffectingCombat("PLAYER")) then
    local mt,mt2 = OW_GetMWDefaults()
    self:SetAttribute("type", "macro")
    self:SetAttribute("macrotext-mwup", mt)
    self:SetAttribute("ctrl-macrotext-mwup", mt)
    self:SetAttribute("alt-macrotext-mwup", mt)
    self:SetAttribute("shift-macrotext-mwup", mt)
    self:SetAttribute("macrotext-mwdn", mt2)
    self:SetAttribute("ctrl-macrotext-mwdn", mt2)
    self:SetAttribute("alt-macrotext-mwdn", mt2)
    self:SetAttribute("shift-macrotext-mwdn", mt2)
  end
end

local secureButton = CreateFrame("Button","OW_BuffFrame",nil,"SecureActionButtonTemplate")
secureButton:RegisterForClicks("AnyUp")
secureButton:SetScript("PreClick",OW_preclick)
secureButton:SetScript("PostClick",OW_postclick)

-- slash command handler
local function OW_slash(opts)
  if not UnitAffectingCombat("PLAYER") then
    if strfind(strupper(opts),"SRESET") then
      for v1 in pairs(OW_NS.situbuffdb) do
        if v1 ~= OW_playername then
          OW_NS.situbuffdb[v1] = {}
        end
      end
      DEFAULT_CHAT_FRAME:AddMessage(OW_NS.sresmsg,1,0,0)
    elseif strfind(strupper((opts or "")),"CHECKBINDINGS") then
      OW_checkbindings()
    elseif strfind(strupper((opts or "")),"MACRO") then
      InterfaceOptionsFrame_OpenToCategory(OW_NS.mwm)
    else
      InterfaceOptionsFrame_OpenToCategory("O-Wheely!") 
    end
  else
    DEFAULT_CHAT_FRAME:AddMessage("Error: /ow is disabled in combat",1,0,0)
  end
end

SLASH_OW1 = "/ow"
SlashCmdList["OW"] = OW_slash
